#!/usr/bin/env python
#-*- coding:utf-8 -*-



 # So that they both go to the same file
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>

# ------------
#| Change Log MrQ BorderBox|
# ------------
# Older version "Create border step by step.py" here: http://gimpchat.com/viewtopic.php?f=9&t=17321
# Rel 1.0: /29-07-2021/ : MareroQ: Initial release with Preview by Claude Lion (cli345) and tabs by J.F. Garcia (arakne)
# Preview by fu_preview (modificed gimpfu) by Claude Lion [aka cli345] http://gimpchat.com/viewtopic.php?f=9&t=19420
# gimfu3 with tabs  https://www.arakne.es/en/gimp-plugin-follow-path-map-values-on-the-y-coordinate/
# Pespective Border Plug-in by Tin Tran https://gimplearn.net/viewtopic.php?f=3&t=2142
# For GIMP 2.10 this plugin requires installation of additional plugins (EXIST IN GIMP-2.8 or from this side: https://www.gimpscripts.net or on http://gimpchat.com/):
# apply-canvas.exe, bump-map.exe, cubism.exe, engrave.exe, felimage-noise.exe, mosaic.exe, noise-generator.exe, noise-spread.exe, pixelize.exe, video.exe, voronoi.exe.
# ------------
# Rel 1.1 /08.08.2021/
# ------------
# - bug fix (for FG/BG stroke colors)
# - added option:
# 1. Create a frame on the layer.
# 2. Create a shadow for the frame (Gegl Drop Shadow).
# 3. Perspective frame in accordance with the Tin Tran code.
# 4. Select the layer mode opacity of the frame.
# ------------
# Rel 1.2  /11.08.2021/
# ------------
# - bug fix: for Perspective frame - replace for Additional Effec (???) to Frame Fill (!!!)
#            for Apply as: On a layer(without Fill Underlay -> now with this option)
# - from Gegl Drop Shadow: - add Grow Shape: " Square "," Circle ", " Diamond "
#                          - create this effect for single layer
# ------------
# Rel 1.3 /15.08.2021/
# ------------
# - from Gegl Drop Shadow: Add to: " No "," To Border ", " To Passe-Partout ", " To Both  "
# - bug fix: for Perspective & Path
# ------------
# Rel 1.4 /19.08.2021/
# ------------
# - Add Tab 'Round Corners' by code Markku Koppinen "round_path_corners.py"
#   Author's website - http://kmarkku.arkku.net/Path_modify_files/Gimp_master.html
#   More information  and discussion at gimp-forum.net: https://www.gimp-forum.net/Thread-Rounding-corners-of-a-straight-edge-path
# - Use 'Corner Type': for grandient & patterns
# - For Fill Passe-Partout add Patterns-desaturate
# ------------
# Rel 1.5 /24.08.2021/
# ------------
# - Add BLURRED Layer for Underlay -> Fill 
# - Add Tab "Distress"
# - Add select colors for Stroke

#####################################################################################################################################################
############################################# START round_path_corners.py ###########################################################################
#####################################################################################################################################################

# Experimental Gimp plugin to round corners of a straight-edge path
#
# History:
# v0.1: 2020-24-05: Plugin "Round corners"
# v0.2: 2020-26-05: Dropped the option for the two-parameter custom shape
# v0.3: 2020-27-06: Bug fix (using selction + anchor outside of canvas)
# v0.4: 2021-28-06: Better wording in the GUI.

# (c) Markku Koppinen 2020, 2021

from __future__ import division, print_function
from math import *
from copy import deepcopy

#==============================================================
#             class BCurve       
#==============================================================
# Note: The plane is treated as the complex plane.

class BCurve(object):
    
    ControlPoint = complex
    
    class GimpVectors(object):

        def __init__(self,
                     stroke_list,            # [GimpStroke]
                     name='GimpVectors'      # string
                     ):
            self.stroke_list = stroke_list
            self.name = name
        def __str__(self):
            s = 'GimpVectors '+self.name
            count = 0
            for stroke in self.stroke_list:
                s += '\n  Stroke '+str(count)+': GimpStroke'
                s += '\n    '+str(stroke)
                count += 1
            return s
        def gv2vectors_object(self, image, name=None):

            def cp_list2xy_list(cp_list):
                xy = []
                for cp in cp_list:
                    xy += [cp.real, cp.imag]
                return xy
            if name is None:
                name = self.name
            vectors = pdb.gimp_vectors_new(image, name)
            for stroke in self.stroke_list:
                closed = stroke.closed
                xy_list = cp_list2xy_list(stroke.cp_list)
                stroke_id = pdb.gimp_vectors_stroke_new_from_points(
                             vectors, 0, len(xy_list), xy_list, closed)
            return vectors
    
    class GimpStroke(object):

        def __init__(self,
                     cp_list,                   # [ControlPoint]
                     closed,                    # boolean
                     stroke_name = 'GimpStroke' # string
                     ):
            self.stroke_name = stroke_name
            self.cp_list = cp_list
            self.closed = closed
        def __str__(self):
            s = self.stroke_name + '; Control points:'
            for cp in self.cp_list:
                s += '\n    '+str(cp)
            s += '\n    closed: '+str(self.closed)
            return s
        def gs2bc(self):

            head = self.cp_list[0]
            tail = self.cp_list[-1]
            inner = self.cp_list[1:-1]
            if len(inner) > 1: # Must be divisible by 4
                ba_list = []
                count = 0
                for i in range(0,len(inner)-1,3):
                    ba_list.append(BCurve.BezierArc(inner[i:i+4], 'arc '+str(count)))
                    count += 1
            elif len(inner) == 1: # Stroke has only one anchor
                ba_list = [BCurve.BezierArc([inner[0]], 'arc 0')]
            else:
                raise Exception("BCurve.GimpStroke.gs2bc: No anchors in stroke?")
            return BCurve.BezierCurve(
                               bezier_arcs = ba_list,
                               head_handle = head,
                               tail_handle = tail,
                               closed = self.closed,
                               curve_name = self.stroke_name
                               )
    
    class BezierCurve(object):

        def __init__(self,
                     bezier_arcs=[],           # [BezierArc]
                     head_handle=None,         # ControlPoint
                     tail_handle=None,         # ControlPoint
                     closed=False,             # boolean
                     curve_name='BezierCurve', # string
                     ):
            self.curve_name = curve_name
            self.bezier_arcs = bezier_arcs
            if head_handle is None:
                self.head_handle = bezier_arcs[0].cp4[0]
            else:
                self.head_handle = head_handle
            if tail_handle is None:
                self.tail_handle = bezier_arcs[-1].cp4[-1]
            else:
                self.tail_handle = tail_handle
            self.closed = closed
        def __str__(self):
            s = self.curve_name + ': Bezier curve' + '; Bezier arcs:'
            count = 0
            for arc in self.bezier_arcs:
                s += '\n  arc '+ str(count)+ ': ' +str(arc)
                count += 1
            s += '\n  Head handle:    '+str(self.head_handle)
            s += '\n  Tail handle:    '+str(self.tail_handle)
            s += '\n  Closed:'    +str(self.closed)
            return s
        def bc2gs(self):

            cp_list = [self.head_handle]
            if len(self.bezier_arcs[0].cp4) > 1:
                for ba in self.bezier_arcs:
                    cp_list += ba.cp4[:3]
                cp_list.append(self.bezier_arcs[-1].cp4[-1])
            else:
                # Only one BezierArc and it has cp4 = [p0]
                # (may rise from Gimp stroke with only one anchor).
                cp_list.append(self.bezier_arcs[0].cp4[0])
            cp_list.append(self.tail_handle)
            return BCurve.GimpStroke(cp_list, self.closed, self.curve_name)
        def bc2vectors_object(self, image, name=None):

            if name is None:
                name = self.curve_name
            gs = self.bc2gs()             # BCurve.GimpStroke
            gv = BCurve.GimpVectors([gs]) # BCurve.GimpVectors
            return gv.gv2vectors_object(image, name)
    
    class BezierArc(object):

        def __init__(self,
                     cp4,                 # cp4=[p0,p1,p2,p3] with pi:ControlPoint
                                          # or [p0]
                     arc_name='BezierArc' # string
                     ):
            self.cp4 = cp4
            self.arc_name = arc_name
        def reverse(self):
            self.cp4.reverse()
        def __str__(self):
            s = 'BezierArc, control points:'
            for cp in self.cp4:
                s += '\n  '+ str(cp)
            return s

# -----------------
# vectors_object2gv
# -----------------

def vectors_object2gv(vectors_object):
    def xy_list2cp_list(xy_list): # xy_list = [x,y,x,y,...,]
        cp = []
        for i in range(0,len(xy_list),2):
            cp.append(BCurve.ControlPoint(xy_list[i], xy_list[i+1]))
        return cp
    stroke_list = []
    count = 0
    for stroke in vectors_object.strokes:
        xy_list = stroke.points[0]
        closed = stroke.points[1]
        cp_list = xy_list2cp_list(xy_list)
        stroke_name = vectors_object.name + ' stroke '+str(count)
        stroke_list.append(BCurve.GimpStroke(cp_list, closed, stroke_name))
        count += 1
    return BCurve.GimpVectors(stroke_list, vectors_object.name)


#==============================================================
#             Auxiliary
#==============================================================

def cross(v,w):
    return (v*w.conjugate()).imag

def collinear(points): # points: [complex] (The plane is the complex plane.)
    ZERO = 1e-5
    n = len(points)
    if n < 3:
        return True
    for i in range(n):
        for j in range(i+1,n):
            for k in range(j+1,n):
                a,b,c = points[i], points[j], points[k]
                norm = max(abs(-a+b),abs(-a+c),abs(-b+c)) ** 2
                if abs(cross(-a+b,-a+c)) > ZERO * norm:
                    #print(str(abs(cross(-a+b,-a+c))/norm)+': NOT collinear: '+str(b))
                    return False
                #else:
                #    print(str(abs(cross(-a+b,-a+c))/norm)+': Collinear: '+str(b))
    return True

# ---------------------------------
# circular_arc_in_tangent_triangle
# ---------------------------------

def circular_arc_in_tangent_triangle(A,B,C):
    a = abs(-A+C)
    c = abs(-A+B)
    h = (4./3)*(c/(2*a+c))
    return [A, (1-h)*A + h*C, (1-h)*B + h*C, B]

#==============================================================
#                       Pre- and postprocessing
#==============================================================

def point_in_selection(x,y,image):
    if (0 <= x < image.width) and (0 <= y < image.height):
        _,pixel = pdb.gimp_drawable_get_pixel(image.selection, x, y)
        return (pixel[0] > 127)
    else:
        return False

# ---------------------
# remove_excess_anchors
# ---------------------

def remove_excess_anchors(cp4list, closed=False):
    if len(cp4list) <= 1:
        return cp4list
    new_cp4list = []
    p0,p1,p2,p3 = cp4list[0]
    for i in range(1,len(cp4list)):
        try_cp4 = cp4list[i]
        if collinear([p0, p3, try_cp4[-1]]):
            p3 = try_cp4[-1]
        else:
            new_cp4list.append([p0,p0,p3,p3])
            p0,p1,p2,p3 = try_cp4
    new_cp4list.append([p0,p0,p3,p3])

    if closed: # Fix the gap-closing arcs:
        p0,p1,p2,p3 = new_cp4list[0] # first cp4
        q0,q1,q2,q3 = new_cp4list[-1] # last cp4
        close_gap_cp4 = [q0,q0,p3,p3]
        new_cp4list[0] = close_gap_cp4
        new_cp4list[-1] = close_gap_cp4

    return new_cp4list

# ---------
# subdivide
# ---------

def subdivide(image, cp4list, simplify_parameters, print_info=False):
    ZERO = 1e-8

    subdivision_anchors = [[False,False] for i in range(len(cp4list))] # Initially all False
    selection_case = simplify_parameters.selection_case
    # Subdivision points: Points in the selection or outside of it.
    if selection_case == 'ignore':
        #return [cp4list]
        pass
    elif selection_case == 'inside':
        # Points not in the selection are made subdivision points.
        for i in range(len(cp4list)):
            cp4 = cp4list[i]
            x,y = cp4[0].real, cp4[0].imag

            if not point_in_selection(x,y,image):
                subdivision_anchors[i][0] = True
            x,y = cp4[-1].real, cp4[-1].imag

            if not point_in_selection(x,y,image):
                subdivision_anchors[i][1] = True
    elif selection_case == 'outside':
        # Points in the selection are made subdivision points.
        for i in range(len(cp4list)):
            cp4 = cp4list[i]
            x,y = cp4[0].real, cp4[0].imag

            if point_in_selection(x,y,image):
                subdivision_anchors[i][0] = True
            x,y = cp4[-1].real, cp4[-1].imag

            if point_in_selection(x,y,image):
                subdivision_anchors[i][1] = True
    else:
        raise Exception("subdivide: Unknown selection_case: "+selection_case)
    # Now the subdivision points are marked. Do the subdivision:
    split_happens = False
    cp4list_list = []
    cumulate_cp4list = []
    for i in range(len(cp4list)):
        cp4 = cp4list[i]
        split_at_start, split_at_end = subdivision_anchors[i]
        if split_at_start and split_at_end:
            split_happens = True
            cp4list_list.append(cumulate_cp4list)
            cp4list_list.append([cp4])
            cumulate_cp4list = []
        elif split_at_start:
            split_happens = True
            cp4list_list.append(cumulate_cp4list)
            cumulate_cp4list = [cp4]
        elif split_at_end:
            split_happens = True
            cp4list_list.append(cumulate_cp4list + [cp4])
            cumulate_cp4list = []
        else:
            cumulate_cp4list.append(cp4)
    if subdivision_anchors[-1][1] == False:
        cp4list_list.append(cumulate_cp4list)
    cp4list_list = [x for x in cp4list_list if len(x) > 0]
    if print_info:
        print("\n\nsubdivide: end points of cp4list:")
        print(str((cp4list[0][0], cp4list[-1][-1])))
        print("subdivide: end points of subdivision:")
        for cp4l in cp4list_list:
            print(str((cp4l[0][0], cp4l[-1][-1])))
    return cp4list_list

# ---------------------
# remove_zero_straights
# ---------------------

def remove_zero_straights(cp4list):
    if len(cp4list) == 1:
        if len(cp4list[0]) == 1:
            return cp4list
    ZERO = 1e-6
    new_cp4list = []
    for i in range(len(cp4list)):
        cp4 = cp4list[i]
        p0,p1,p2,p3 = cp4
        if not ((abs(-p0+p3) < ZERO) and
               (abs(-p0+p1) < ZERO) and
               (abs(-p2+p3) < ZERO)):
            new_cp4list.append(cp4)
            continue
        try:
            prev = new_cp4list[-1]
            prev[-1] = p0
        except IndexError:
            continue
    return new_cp4list

#==============================================================
#                           Drawing
#==============================================================
# ------------------------
# gimp_draw_vectors_object
# ------------------------

def gimp_draw_vectors_object(image, vectors_object, visible=True):
    # Insert a Gimp vectors object in Gimp.
    # Adapted from http://registry.gimp.org/files/lqr_wpset-0.14.py.txt
    # by Mike Kazantsev.
    def gimp_insert_vectors(image, vectors_object, position=0):
        if gimp.version >= (2, 8, 0):
            # 2.7.0 has gimp-image-insert-vectors, but fails with
            # parent=None
            parent = None
            return pdb.gimp_image_insert_vectors(
                     image, vectors_object, parent, position)
        else:
            return pdb.gimp_image_add_vectors(
                     image, vectors_object, position)# Deprecated!
    gimp_insert_vectors(image, vectors_object)
    vectors_object.visible = visible
    return vectors_object

# For debugging and experimenting only
def draw_cp4list(image, cp4list, name='cp4list'):
    balist = [BCurve.BezierArc(cp4=cp4) for cp4 in cp4list]
    bc = BCurve.BezierCurve(bezier_arcs = balist)
    vectors = bc.bc2vectors_object(image, name=name)
    gimp_draw_vectors_object(image, vectors, visible=True)

#==============================================================
#             Round vertices
#==============================================================

class ParameterRecord(object):
    def __init__(self, distance, custom1=None):#, custom2=None):
        self.distance = distance
        self.custom1 = custom1
        #self.custom2 = custom2

class RoundingParameters(object):
    def __init__(self,
                 parameters,   # ParameterRecord
                 shape_case,
                 arc_end_case,
                 selection_case
                 ):
        self.parameters = parameters
        self.shape_case = shape_case
        self.arc_end_case=arc_end_case,
        self.selection_case = selection_case
    def __str__(self):
        s = 'RoundingParameters:'
        s += '\n  parameters = '+str(self.parameters)
        s += '\n  shape_case = '+str(self.shape_case)
        s += '\n  arc_end_case = '+str(self.arc_end_case)
        s += '\n  selection_case = '+str(self.selection_case)
        return s

# ---------------
# round_cp4_cp4_1
# ---------------

def round_cp4_cp4_1(left_cp4, right_cp4, parameters):
    dist, ratio = parameters
    p0,p1,p2,p3 = left_cp4
    q0,q1,q2,q3 = right_cp4
    if dist == None: # Means that shape_end_case = 'midpoints'
        d_left = abs(-p0+p3)/2
        d_right = abs(-q0+q3)/2
    else:
        d_left = min(dist, abs(-p0+p3)/2)
        d_right = min(dist, abs(-q0+q3)/2)
    R0 = p3 + d_left * (-p3+p0)/abs(-p3+p0)
    R3 = q0 + d_right * (-q0+q3)/abs(-q0+q3)
    R1 = (1-ratio) * R0 + ratio * p3
    R2 = (1-ratio) * R3 + ratio * p3
    return [[R0,R1,R2,R3]]

# ---------------
# round_cp4_cp4_2
# ---------------

def round_cp4_cp4_2(left_cp4, right_cp4, parameters):
    dist, ratio1, ratio2 = parameters
    p0,p1,p2,p3 = left_cp4
    q0,q1,q2,q3 = right_cp4
    if dist == None: # Means that shape_end_case = 'midpoints'
        d_left = abs(-p0+p3)/2
        d_right = abs(-q0+q3)/2
    else:
        d_left = min(dist, abs(-p0+p3)/2)
        d_right = min(dist, abs(-q0+q3)/2)
    P0 = p3 + d_left * (-p3+p0)/abs(-p3+p0)
    Q3 = q0 + d_right * (-q0+q3)/abs(-q0+q3)
    P2 = (1-ratio1) * P0 + ratio1 * p3
    Q1 = (1-ratio1) * Q3 + ratio1 * q0
    P3 = Q0 = (P2 + Q1) / 2
    P1 = (1-ratio2) * P0 + ratio2 * P2
    Q2 = (1-ratio2) * Q3 + ratio2 * Q1
    return [[P0,P1,P2,P3], [Q0,Q1,Q2,Q3]]

# --------------------
# round_cp4_cp4_circle
# --------------------

def round_cp4_cp4_circle(left_cp4, right_cp4, parameters):
    p0,p1,p2,p3 = left_cp4
    q0,q1,q2,q3 = right_cp4
    dist = parameters[0]
    if dist is None: # Means that shape_end_case = 'midpoints'
        d = min(abs(-p0+p3)/2, abs(-q0+q3)/2)
    else:
        d = min(dist, abs(-p0+p3)/2, abs(-q0+q3)/2)
    R0 = p3 + d * (-p3+p0)/abs(-p3+p0)
    R3 = q0 + d * (-q0+q3)/abs(-q0+q3)
    return [circular_arc_in_tangent_triangle(R0,R3,p3)]

# ---------------------
# round_corners_cp4list
# ---------------------

def round_corners_cp4list(cp4list, rp):
    if rp.shape_case == 'straight':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, 0]   # [distance, 0]
    elif rp.shape_case == 'circular':
        corner_rounding = round_cp4_cp4_circle
        parameters = [rp.parameters.distance]      # [distance]
    elif rp.shape_case == 'parabola':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, 2/3] # [distance, 2/3]
    elif rp.shape_case == 'maximal':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, 1]   # [distance, 1]
    elif rp.shape_case == 'through':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, 4/3] # [distance, 3/4]
    elif rp.shape_case == 'cusp':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, 2]   # [distance, 2]
    elif rp.shape_case == '1 custom':
        corner_rounding = round_cp4_cp4_1
        parameters = [rp.parameters.distance, rp.parameters.custom1] # [distance, custom1]
    elif rp.shape_case == '2 custom':
        corner_rounding = round_cp4_cp4_2
        parameters = [rp.parameters.distance,      # [distance, custom1, custom2]
                      rp.parameters.custom1,
                      rp.parameters.custom2]
    if len(cp4list) <= 1:
        return deepcopy(cp4list)
    corner_cp4lists = [] # [[cp4,cp4]]
    for i in range(len(cp4list)-1):
        left, right = cp4list[i:i+2]
        corner_cp4lists.append(corner_rounding(left, right, parameters))
    # Join the corner arcs with straight line segments:
    new_cp4list = []
    for i in range(len(corner_cp4lists)-1):
        corner1_cp4list = corner_cp4lists[i]   # [cp4,cp4]
        corner2_cp4list = corner_cp4lists[i+1] # [cp4,cp4]
        arc1 = corner1_cp4list[-1] # cp4
        arc2 = corner2_cp4list[0] # cp4
        segment = [arc1[-1], arc1[-1], arc2[0], arc2[0]]
        new_cp4list += corner1_cp4list
        new_cp4list.append(segment)
    new_cp4list += corner_cp4lists[-1]
    #
    start_cp4 = cp4list[0]
    start_arc = corner_cp4lists[0][0]
    p0,p1,p2,p3 = start_arc
    new_start_arc = start_cp4[:2] + [p0, p0]
    #
    end_cp4 = cp4list[-1]
    end_arc = corner_cp4lists[-1][-1]
    p0,p1,p2,p3 = end_arc
    new_end_arc =  [p3, p3] + end_cp4[-2:]
    #
    new_cp4list = [new_start_arc] + new_cp4list + [new_end_arc]
    return new_cp4list

# --------------------
# round_corners_bcurve
# --------------------

def round_corners_bcurve(image, bc, rp, remove_excess):
    cp4list = [ba.cp4 for ba in bc.bezier_arcs]
    if bc.closed: # Close gap twice: add one arc at the head and at the tail:
        start = cp4list[0][0]
        end = cp4list[-1][-1]
        head_cp4 = tail_cp4 = [end, end, start, start]
        cp4list = [head_cp4] + cp4list + [tail_cp4]
    # --- Preprocessing ---
    list_cp4lists = subdivide(image, cp4list, rp, print_info=False)

    closed = bc.closed and (len(list_cp4lists) == 1)
    if remove_excess:
        rem_list_cp4lists = []
        for cp4l in list_cp4lists:
            rem_list_cp4lists.append(remove_excess_anchors(cp4l, closed))
        list_cp4lists = rem_list_cp4lists
    # --- Rounding ---
    new_cp4list = []
    for one_cp4list in list_cp4lists:
        new_cp4list += round_corners_cp4list(one_cp4list, rp) # Main call
    if bc.closed: # Remove added head and tail
        new_cp4list = new_cp4list[1:-1]
    # --- Postprocessing ---
    new_cp4list = remove_zero_straights(new_cp4list)
    new_balist = [BCurve.BezierArc(cp4=cp4) for cp4 in new_cp4list]
    return BCurve.BezierCurve(bezier_arcs = new_balist,
                              closed=bc.closed,
                              curve_name='rounded bc',
                              )

# ------------------
# round_corners_path
# ------------------

def round_corners_path(image, path, rp, remove_excess):
    gv = vectors_object2gv(path) # BCurve.GimpVectors
    new_stroke_list = []
    for gs in gv.stroke_list: # BCurve.GimpStroke
        bc = gs.gs2bc()       # BCurve.BezierCurve
        new_bc = round_corners_bcurve(image, bc, rp, remove_excess) # Main call
        new_gs = new_bc.bc2gs()
        new_stroke_list.append(new_gs)
    new_gv = BCurve.GimpVectors(stroke_list = new_stroke_list)
    if rp.shape_case not in ('1 custom','2 custom'):
        name = path.name+'|'+rp.shape_case
    elif rp.shape_case == '1 custom':
        name = path.name+'|custom: '
        name += str(rp.parameters.custom1)
    elif rp.shape_case == '2 custom':
        name = path.name+'|custom: '
        name += str(rp.parameters.custom1)
        name += ', '
        name += str(rp.parameters.custom2)
    return new_gv.gv2vectors_object(image, name=name)

selection_options = [ # (description, identifier)
        ('Ignore any selection',                                      'ignore'),
        ('Round the corners only inside the selection', 'inside'),
        ('Round the corners only outside of the selection', 'outside'),
        ]

shape_options = [ # (description, identifier)
        ('circular',                   'circular'),
        ('straight (parameter = 0)',   'straight'),
        ('parabola (2/3)',             'parabola'),
        ('close to vertex (1)',        'maximal'),
        ('through vertex (4/3)',       'through'),
        ('cusp (2)',                   'cusp'),
        ('one-parametric custom shape - set parameter value below',             '1 custom'),
        ]

arc_end_options = [ # (description, identifier)
        ('Use fixed distance from vertices - give the distance below', 'fixed'),
        ('Use midpoints of edges (as far as possible)',                                     'midpoints'),
        ]

#==============================================================
#             Main procedure
#==============================================================

def round_corners(image,
                       path, 
                       #remove_excess,   # boolean
                       arc_end_option, 	 # integer
                       distance,         # float
                       shape_option,     # integer
                       custom1,          # float
                       #custom2,         # float
                       selection_option, # integer
                       ):
    remove_excess = True # No option for the user!
    selection_case = selection_options[selection_option][1]
    shape_case = shape_options[shape_option][1]
    arc_end_case = arc_end_options[arc_end_option][1]
    if (arc_end_case == 'midpoints'):
        distance = None # This will indicate the case.
    rp = RoundingParameters(
                            parameters=ParameterRecord(distance, custom1),#, custom2),
                            shape_case=shape_case,
                            arc_end_case=arc_end_case,
                            selection_case=selection_case
                            )
    path_rounded = round_corners_path(image, path, rp, remove_excess) # Main call
    pdb.gimp_image_undo_group_start(image)
    gimp_draw_vectors_object(image, path_rounded, visible=True)
    pdb.gimp_image_undo_group_end(image)
    return path_rounded

#####################################################################################################################################################
############################################# END round_path_corners.py #############################################################################
#####################################################################################################################################################


#####################################################################################################################################################
############################################# START fu_preview 0.1.1 by Claude Lion $ J.F. Garcia ###################################################
#####################################################################################################################################################

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb
import os, time, sys
import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}


def register(proc_name, blurb, help, author, copyright, date, label,
             imagetypes, params, results, function,
             menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
                return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements "
                          "(%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead"
                       % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright,
                                       date, label, imagetypes,
                                       plugin_type, params, results,
                                       function, menu, domain,
                                       on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date,
         label, imagetypes, plugin_type,
         params, results, function, menu, domain,
         on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]],
                     x[1],
                     _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode",
                                     "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright,
                               date, label, imagetypes, plugin_type,
                               params, results)

        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date,
     label, imagetypes, plugin_type,
     params, results, function, menu, domain,
     on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
#    import pango
    gimpui.gimp_ui_init () # add MrQ
    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            self.set_activates_default(True)
            #~ self.connect('paste-clipboard', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('inserted-text', lambda _1, _2, _3, _4: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('deleted-text', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate()) # modif C.L

        def get_value(self):
            return self.get_text()


    class LabelTitle(gtk.Label):
        #def __init__ (self, default="", **kwargs): # modif C.L. ???
        def __init__(self, default="", align = (0.0, 0.0)):		
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default="", **kwargs): # modif C.L.
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))
            self.buffer.connect('changed', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.			

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            # modif C.L.
            def react(adj):
                return self.get_toplevel().la_previ.invalidate()
            #~ self.adj.connect('value-changed', lambda adj: self.get_toplevel().la_previ.invalidate())
            self.adj.connect('value-changed', react)
            # end modif C.L.
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=10, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            self.adj.connect('value-changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))


    class ToggleEntry(gtk.ToggleButton):
        #~ def __init__(self, default=0):
        def __init__(self, default=0,labels=[_("No"),_("Yes")], **kwargs): # modif C.L.
            
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()

            #~ self.connect("toggled", self.changed)
            # modif C.L.
            def react(tog):
                self.changed(tog)
                return kwargs['dialg'].la_previ.invalidate()
            self.connect("toggled", react)
            # end modif C.L.


            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()



    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0)), **kwargs): # modif C.L.
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=(), **kwargs): # modif C.L.
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)
            self.connect('changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.

        def get_value(self):
            return self.get_active()

    def FileSelector(default="", title=None, **kwargs): # modif C.L.
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            if default == "/": default = ""
            return DirnameSelector(default)
        else:
            return FilenameSelector(default, title=title, save_mode=False)

    class FilenameSelector(gtk.HBox):
        #gimpfu.FileChooserButton
        def __init__(self, default, save_mode=True, title=None, **kwargs): # modif C.L.
            super(FilenameSelector, self).__init__()
            if not title:
                self.title = _("Python-Fu File Selection")
            else:
                self.title = title
            self.save_mode = save_mode
            box = self
            self.entry = gtk.Entry()
            image = gtk.Image()
            image.set_from_stock(gtk.STOCK_FILE, gtk.ICON_SIZE_BUTTON)
            self.button = gtk.Button()
            self.button.set_image(image)
            box.pack_start(self.entry)
            box.pack_start(self.button, expand=False)
            self.button.connect("clicked", self.pick_file)
            if default:
                self.entry.set_text(default)

        def show(self):
            super(FilenameSelector, self).show()
            self.button.show()
            self.entry.show()

        def pick_file(self, widget):
            entry = self.entry
            dialog = gtk.FileChooserDialog(
                         title=self.title,
                         action=(gtk.FILE_CHOOSER_ACTION_SAVE
                                     if self.save_mode else
                                 gtk.FILE_CHOOSER_ACTION_OPEN),
                         buttons=(gtk.STOCK_CANCEL,
                                gtk.RESPONSE_CANCEL,
                                gtk.STOCK_SAVE
                                     if self.save_mode else
                                gtk.STOCK_OPEN,
                                gtk.RESPONSE_OK)
                        )
            dialog.set_alternative_button_order ((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))
            dialog.show_all()
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                entry.set_text(dialog.get_filename())
            dialog.destroy()

        def get_value(self):
            return self.entry.get_text()


    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.FileChooserButton.__init__(self,
                                           _("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    # modif C.L.
    class ColorSelector(gimpui.ColorSelector):
        def __init__(self, default=gimpcolor.RGB(1.0, 0, 0), **kwargs):
            ret = gimpui.ColorSelector.__init__(self, default)
            self.connect('color-changed', lambda _: kwargs['dialg'].la_previ.invalidate())
            return ret
    class FontSelector(gimpui.FontSelector):
        def __init__(self, default="Sans", **kwargs):
            ret = gimpui.FontSelector.__init__(self, default)
            self.connect('font-set', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PatternSelector(gimpui.PatternSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PatternSelector.__init__(self, default)
            self.connect('pattern-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class GradientSelector(gimpui.GradientSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.GradientSelector.__init__(self, default)
            self.connect('gradient-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PaletteSelector(gimpui.PaletteSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PaletteSelector.__init__(self, default)
            self.connect('palette-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    # end modif C.L.
    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : ColorSelector, # modif C.L.
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : PatternSelector, # modif C.L.
            PF_GRADIENT    : GradientSelector, # modif C.L.
            PF_PALETTE     : PaletteSelector, # modif C.L.
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    # modif C.L.
    hboxPrincipal = gtk.HBox(spacing=10)
	
    #dialog.vbox.pack_start(hboxPrincipal)			        # modif MrQ
    swin = gtk.ScrolledWindow()                             # New line MrQ
    swin.add_with_viewport(hboxPrincipal)                   # New Line MrQ
    dialog.vbox.pack_start(swin)                            # New Line MrQ
    swin.show()                                             # New Line MrQ	

    hboxPrincipal.show()
    
    dialog.la_previ = gimpui.DrawablePreview(start_params[1])
    dialog.la_previ.set_size_request(400,400)               # New Line MrQ	
    dialog.la_previ.la_img = start_params[0]
    dialog.la_previ.la_layer = start_params[1]
    hboxPrincipal.pack_start(dialog.la_previ)

    dialog.set_default_size(1100, 600)  					# New Line MrQ
    dialog.la_previ.show()
    # end modif C.L.



    vbox = gtk.VBox(False, 2)
    vbox.set_border_width(2)
    vbox.set_size_request(480, 0)                           # modif C.L. + MrQ
    #~ dialog.vbox.pack_start(vbox)
    hboxPrincipal.pack_start(vbox) # modif C.L.
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})
    # modif C.L.
    def ifinvalidated(p_la_previ):
        
        # get dimension
        #===============
        x, y = p_la_previ.get_position()
        w, h = p_la_previ.get_size()
        
        # Effet
        #=========
        # has alpha
        lb_alpha = pdb.gimp_drawable_has_alpha(p_la_previ.la_layer)
        
        # img, layer temporaires
        tempImg = pdb.gimp_image_new(w, h, 0)
        #~ tempLayer = pdb.gimp_layer_new(tempImg, w, h, 0, 'temp', 100, 0)
        tempLayer = pdb.gimp_layer_new(tempImg, w, h, lb_alpha, 'temp', 100, 0)
        pdb.gimp_image_add_layer(tempImg, tempLayer, 0)
        
        # copie orig dans tempo
        rgnPreviIn = p_la_previ.la_layer.get_pixel_rgn(x, y, w, h)
        rgnTemp = tempLayer.get_pixel_rgn(0, 0, w, h)
        rgnTemp[0:w, 0:h] = rgnPreviIn[x:x+w, y:y+h]
        
        # effect
        tempImg.undo_group_start()
        tempImg.disable_undo()
        
        params = []

        try:
            for wid in edit_wids:
                params.append(wid.get_value())
        except EntryValueError:
            warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
        else:
            try:
                paramsTot = (tempImg, tempLayer) + tuple(params)
                _set_defaults(proc_name, start_params + tuple(params))
                function(*paramsTot)
                
                res = tempImg.active_layer
                
                if not lb_alpha: pdb.gimp_layer_flatten(res)
                
                tempImg.undo_group_end()
                
                rgnRes = res.get_pixel_rgn(0, 0, w, h)
                
                # mise dans previ
                #================
                p_la_previ.draw_buffer(rgnRes[0:w, 0:h], int(3+lb_alpha)*w)
                
                pdb.gimp_image_delete(tempImg)
                
                #pdb.gimp_display_new(tempImg)
            except CancelError:
                pass
            except Exception as e:
                dialog.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                error_dialog(dialog, proc_name)
                raise
    # end modif C.L.
    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()
    dialog.la_previ.connect('invalidated', ifinvalidated) # modif C.L.
    dialog.connect("response", response)
    iniX=0
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]

        tab, incr, rows, connect, cols = (0, 1, 1, None, 1)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE:
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW):
                wid = _edit_mapping[pf_type](def_val, params[i][4], dialg=dialog) # modif C.L.
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4],dialg=dialog)
                else:
                    wid = _edit_mapping[pf_type](def_val)				


            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val, dialg=dialog) # modif C.L.
            if pf_type!=PF_TITLE:
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', desc)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()
	

	
    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break

        min_args = i

    if len(script_params) > min_args:
        start_params = params[:min_args + 1]

        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE

    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])

    gimp.displays_flush()

    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message
	
	
#####################################################################################################################################################
############################################# END fu_preview 0.1.1 by Claude Lion $ J.F. Garcia #####################################################
#####################################################################################################################################################

####################################################################################################################
################# START gegl_command by cli345 #####################################################################
####################################################################################################################

from sys import platform
from ctypes import *

def load_library (library_name):
	if platform == "linux" or platform == "linux2":
		library_name = library_name + '.so.0'
	elif platform == "win32":
		from ctypes.util import find_library
		library_name = find_library (library_name + "-0")
	else:
		raise BaseException ("TODO")
	return CDLL (library_name)

gimpC = load_library ('libgimp-2.0')


def gegl_command_preview(image, drawable, gegl_graph_string):
	
	class GeglBuffer(Structure):
		pass
	drawable_id = drawable.ID
	
	gegl = load_library ('libgegl-0.4')
	gegl.gegl_init (None, None)
	gimpC.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
	gimpC.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

	x = c_int (); y = c_int (); w = c_int (); h = c_int ()
	non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
	args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
	
	if non_empty:
		source = gimpC.gimp_drawable_get_buffer (drawable_id)
		target = gimpC.gimp_drawable_get_shadow_buffer (drawable_id)
		gegl.gegl_render_op (source, target, "gegl:gegl", *args)
		gegl.gegl_buffer_flush (target)
		gimpC.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
		gimpC.gimp_drawable_update (drawable_id, x, y, w, h)
		gimpC.gimp_displays_flush ()

####################################################################################################
################# END gegl_command #################################################################
####################################################################################################

def rgb_to_hex(rgb_color):
	r = rgb_color[0]
	g = rgb_color[1]
	b = rgb_color[2]
	return '#%02x%02x%02x' % (r,g,b)

def newline(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5):
    return [x1,y1]*3+[x2,y2]*3+[x3,y3]*3+[x4,y4]*3+[x5,y5]*3
	
def distress(image, layer, select,threshold,spread,granularity,smooth,col) :
	new_vectors_d = pdb.gimp_vectors_new(image,"Border for distress")
	pdb.gimp_image_insert_vectors(image, new_vectors_d, None, -1)
	wl = pdb.gimp_image_width(image)
	hl = pdb.gimp_image_height(image)    
	d=select
	pdb.gimp_vectors_stroke_new_from_points(new_vectors_d, 0, 30, newline( d, d, wl-d,d,  wl-d,hl-d, d,hl-d, d,d), 1)
	pdb.gimp_image_select_item(image, 2, new_vectors_d)

	theLayer = pdb.gimp_layer_new(image, wl, hl, 1, "Distress Layer", 100, 0)
	pdb.gimp_image_insert_layer(image,theLayer,None,0)
	pdb.gimp_drawable_edit_fill(theLayer, 2)
	pdb.gimp_selection_invert(image)
	pdb.gimp_drawable_edit_clear(theLayer)
	pdb.gimp_selection_invert(image)   
	pdb.gimp_selection_none(image)
	pdb.gimp_layer_scale(theLayer, wl/granularity, hl/granularity, 1)
	pdb.plug_in_spread(image, theLayer, spread, spread)
	pdb.plug_in_gauss_iir(image, theLayer, smooth, 0, 0)
	pdb.gimp_layer_scale(theLayer, wl, hl, 1)
	pdb.plug_in_threshold_alpha(image, theLayer, threshold)
	pdb.plug_in_gauss_iir(image, theLayer, 1, 1, 1)
	pdb.gimp_image_select_item(image, 2, theLayer)
	pdb.gimp_image_lower_layer_to_bottom(image, theLayer)
	pdb.gimp_selection_invert(image)
	pdb.gimp_drawable_edit_clear(layer)
	pdb.gimp_selection_none(image)
	pdb.gimp_context_push()
	pdb.gimp_context_set_foreground(col)
	pdb.gimp_drawable_edit_fill(theLayer, 0)
	pdb.gimp_context_pop()		
	
def MrQ_BorderBox(image, layer, background,background_col, frame, sel_col,pp,pp_col,c_type, gradient_type, color_fg,color_bg,swap, efect, modepat, opacity,scale_p,merge,
	use_shape,shape_option,distance,
	stroke_tx,stroke_out,color_out,stroke_in,color_in,stroke_layer,color_sl,del_path,
	drop_shadow, shadow_offset_x, shadow_offset_y, shadow_blur, grow_shape, grow_radius, color_s, shadow_opacity,
	use_distress,select,threshold,spread,granularity,smooth,color_d,
	hlp, ptab):

	pdb.gimp_image_undo_group_start(image)
	pdb.gimp_context_push()
	pdb.gimp_selection_none(image)
	
##################### add for "Preview" ###############################################
	
	active_layer = pdb.gimp_image_get_active_layer(image)
	
	for layer in image.layers:
		pdb.gimp_image_set_active_layer(image,layer)
		pdb.gimp_item_set_visible(layer, 0)
	
#######################################################################################	

	black = (0,0,0)
	white = (255,255,255)
	
	gimp.set_foreground(color_fg)
	gimp.set_background(color_bg)
	
	# Swap the current GIMP foreground and background colors
	if swap==1:
		pdb.gimp_context_swap_colors()
		gimp.get_foreground()
		gimp.get_background()	

	copy = pdb.gimp_layer_new_from_drawable(layer, image)
	image.add_layer(copy, 0)
	pdb.gimp_item_set_visible(copy, 1)
	
	b = frame
	e = background
	d = b + e
	#Passe Partout
	if pp>0 and scale_p!=0:		

		passe_partout = pdb.gimp_layer_new(image,image.width+2*pp,image.height+2*pp,RGBA_IMAGE,"Passe Partout",100,0)
		pdb.gimp_image_insert_layer(image,passe_partout,None,0)
		pdb.gimp_layer_set_offsets(passe_partout, -pp, -pp)
		pdb.gimp_image_resize(image,image.width+2*pp,image.height+2*pp, pp, pp)
		pdb.gimp_layer_resize_to_image_size(passe_partout)
		pdb.gimp_edit_clear(passe_partout)			
		pdb.gimp_item_set_name(passe_partout, "Passe Partout")

		pdb.gimp_selection_none(image)	
	
		if pp_col==0:
			pdb.gimp_drawable_fill(passe_partout, 0)
		if pp_col==1:
			pdb.gimp_drawable_fill(passe_partout, 1)
		if pp_col==2:
			pdb.gimp_drawable_fill(passe_partout, 2)
		if pp_col==3:
			pdb.gimp_drawable_fill(passe_partout, 3)
		if pp_col==4:
			pdb.gimp_drawable_fill(passe_partout, 4)
		if pp_col==5:
			pdb.gimp_drawable_fill(passe_partout, 4)
			pdb.gimp_desaturate(passe_partout)

			
	wi = pdb.gimp_image_width(image)
	hi = pdb.gimp_image_height(image)
	
	if scale_p==0:
		new_width = wi + 2*e
		new_height = hi + 2*e
		
	else:
		new_width = wi + 2*d
		new_height = hi + 2*d
	
	wh = pdb.gimp_drawable_width(copy)
	hh = pdb.gimp_drawable_height(copy)

	sw=(wi-wh)/2
	sh=(hi-hh)/2		
	pdb.gimp_layer_set_offsets(copy, sw, sh)


	active_layer = pdb.gimp_layer_new_from_drawable(copy, image)
	pdb.gimp_image_insert_layer(image, active_layer, None, 0)
	active_layer.name = "Layer Copy"
	
	
	if scale_p==0:
		pdb.gimp_image_resize(image,new_width,new_height,e, e)
		
	else:
		pdb.gimp_image_resize(image,new_width,new_height,d, d)	
	if background_col!=5:	
		pdb.gimp_layer_resize_to_image_size(active_layer)
		
	background_layer = pdb.gimp_layer_new_from_drawable(active_layer, image)
	pdb.gimp_image_insert_layer(image, background_layer, None, 0)
	pdb.gimp_item_set_name(background_layer, "Underlay")
	pdb.gimp_item_set_visible(background_layer, 1)
	if background_col==5:
		pdb.gimp_layer_scale(background_layer, new_width,new_height, 1)
		pdb.gimp_layer_resize_to_image_size(background_layer)
		gegl_command_preview(image, background_layer,"focus-blur")
	else:
		pdb.gimp_edit_clear(background_layer)
	pdb.gimp_image_lower_item_to_bottom(image, background_layer)
	
	if pp>0 and scale_p==0:

		passe_partout = pdb.gimp_layer_new(image,layer.width-2*b,layer.height-2*b,RGBA_IMAGE,"Passe Partout",100,0)
		pdb.gimp_image_insert_layer(image,passe_partout,None,0)
		pdb.gimp_layer_set_offsets(passe_partout, b+e, b+e)
		
		if pp_col==0:
			pdb.gimp_drawable_fill(passe_partout, 0)
		if pp_col==1:
			pdb.gimp_drawable_fill(passe_partout, 1)
		if pp_col==2:
			pdb.gimp_drawable_fill(passe_partout, 2)
		if pp_col==3:
			pdb.gimp_drawable_fill(passe_partout, 3)
		if pp_col==4:
			pdb.gimp_drawable_fill(passe_partout, 4)
		if pp_col==5:
			pdb.gimp_drawable_fill(passe_partout, 4)
			pdb.gimp_desaturate(passe_partout)
		
	if e>0 or scale_p==0:
		pdb.gimp_selection_none(image)	
		if background_col==0:
			pdb.gimp_drawable_fill(background_layer, 0)
		if background_col==1:
			pdb.gimp_drawable_fill(background_layer, 1)
		if background_col==2:
			pdb.gimp_drawable_fill(background_layer, 2)
		if background_col==3:
			pdb.gimp_drawable_fill(background_layer, 3)
		if background_col==4:
			pdb.gimp_drawable_fill(background_layer, 4)		

	if gradient_type==0:
		gradtype=0
	if gradient_type==1:
		gradtype=6
	if gradient_type==2:
		gradtype=7
	if gradient_type==3:
		gradtype=8
	reverse_grad = pdb.gimp_context_get_gradient_reverse()

		
	if c_type==0:	# slanted

		new_vectors_bl = pdb.gimp_vectors_new(image,"Border-L")
		pdb.gimp_image_insert_vectors(image, new_vectors_bl, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bl, 0, 30, newline( d, d, d,hi+e-b,  e,hi+e, e,e, d,d), 1)
		else:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bl, 0, 30, newline( d, d, d,hi+d,  e,hi+d+b, e,e, d,d), 1)
		pdb.gimp_item_set_visible(new_vectors_bl, False)
		
		new_vectors_bt = pdb.gimp_vectors_new(image,"Border-T")
		pdb.gimp_image_insert_vectors(image, new_vectors_bt, None, -1)
		if scale_p==0:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bt, 0, 30, newline(e,e,   d,d, wi+e-b,d ,  wi+e,e,    e,e), 1)
		else:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bt, 0, 30, newline(e,e,   d,d,  wi+d,d,  wi+d+b,e,  e,e), 1)
		pdb.gimp_item_set_visible(new_vectors_bt, False)

		new_vectors_br = pdb.gimp_vectors_new(image,"Border-R")
		pdb.gimp_image_insert_vectors(image, new_vectors_br, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_br, 0, 30, newline(wi+e,e,  wi+e,hi+e,   wi+e-b,hi+e-b,  wi+e-b,d,  wi+e,e,), 1)		

		else:			
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_br, 0, 30, newline( wi+d,d,  wi+d+b,e,  wi+d+b,hi+d+b,  wi+d,hi+d,  wi+d,d), 1)
		pdb.gimp_item_set_visible(new_vectors_br, False)
		
		new_vectors_bb = pdb.gimp_vectors_new(image,"Border-B")
		pdb.gimp_image_insert_vectors(image, new_vectors_bb, None, -1)
		if scale_p==0:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bb, 0, 30, newline(d,hi+e-b, wi+e-b,hi+e-b, wi+e,hi+e,  e,hi+e, d,hi+e-b,), 1)
		else:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bb, 0, 30, newline(wi+d+b,hi+d+b, wi+d,hi+d,  d,hi+d,  e,hi+d+b,   wi+d+b,hi+d+b), 1)
		pdb.gimp_item_set_visible(new_vectors_bb, False)
		
	else :	# rectangular top, rectangular side, rectangular mixed
		new_vectors_bl = pdb.gimp_vectors_new(image,"Border-L")
		pdb.gimp_image_insert_vectors(image, new_vectors_bl, None, -1)
		if scale_p==0:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bl, 0, 30, newline( d, e, d,hi+e,  e,hi+e, e,e, d,e), 1)
		else:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bl, 0, 30, newline( d, e, d,hi+2*b+e,  e,hi+2*b+e, e,e, d,e), 1)
		pdb.gimp_item_set_visible(new_vectors_bl, False)
		
		new_vectors_bt = pdb.gimp_vectors_new(image,"Border-T")
		pdb.gimp_image_insert_vectors(image, new_vectors_bt, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bt, 0, 30, newline(e,e,   e,d,  wi+e,d,   wi+e,e,    e,e), 1)
		else:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bt, 0, 30, newline(e,e,   e,d,  wi+2*b+e,d,  wi+2*b+e,e,  e,e), 1)
		pdb.gimp_item_set_visible(new_vectors_bt, False)
		

		new_vectors_br = pdb.gimp_vectors_new(image,"Border-R")
		pdb.gimp_image_insert_vectors(image, new_vectors_br, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_br, 0, 30, newline( wi+e-b,e,  wi+e,e,  wi+e,hi+e,  wi+e-b,hi+e,  wi+e-b,e), 1)
		else:		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_br, 0, 30, newline( wi+d,e,  wi+2*b+e,e,  wi+2*b+e,hi+2*b+e,  wi+d,hi+2*b+e,  wi+d,e), 1)
		pdb.gimp_item_set_visible(new_vectors_br, False)
		
		
		new_vectors_bb = pdb.gimp_vectors_new(image,"Border-B")
		pdb.gimp_image_insert_vectors(image, new_vectors_bb, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bb, 0, 30, newline( e,hi+e, wi+e,hi+e,  wi+e,hi+e-b, e,hi+e-b,  e,hi+e, ), 1)
		else:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_bb, 0, 30, newline(wi+2*b+e,hi+2*b+e, wi+2*b+e,hi+d,  e,hi+d,  e,hi+2*b+e,   wi+2*b+e,hi+2*b+e), 1)
		pdb.gimp_item_set_visible(new_vectors_bb, False)

	
	e_layer = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"Border layer",100,0)
	pdb.gimp_image_insert_layer(image,e_layer,None,0)

	pdb.gimp_selection_none(image)
	

	if sel_col==0:
		pdb.gimp_context_get_foreground()
		pdb.gimp_edit_fill(e_layer,0)
		
	if sel_col==1:
		pdb.gimp_context_get_background()		
		pdb.gimp_edit_fill(e_layer,1)
		
	if sel_col==2:		
		pdb.gimp_context_set_foreground(black)
		pdb.gimp_edit_fill(e_layer,0)
		
	if sel_col==3:		
		pdb.gimp_context_set_foreground(white)			
		pdb.gimp_edit_fill(e_layer,0)

	if sel_col==4 and use_shape!=1:
	
		pattern_layer_bt = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"Texture layer",100,0)
		pdb.gimp_image_insert_layer(image,pattern_layer_bt,None,0)

		pdb.gimp_image_select_item(image, 2, new_vectors_bt)
		pdb.gimp_context_get_pattern()
		pdb.gimp_drawable_fill(pattern_layer_bt, 4)
		pdb.gimp_selection_invert(image)
		pdb.gimp_edit_clear(pattern_layer_bt)
		pdb.gimp_selection_none(image)
		layer_copy = pdb.gimp_layer_new_from_drawable(pattern_layer_bt, image)
		pdb.gimp_image_insert_layer(image,layer_copy,None,0)
		pdb.gimp_item_transform_flip_simple(layer_copy, 1, 1, 0)
		mer = pdb.gimp_image_merge_down(image,layer_copy,EXPAND_AS_NECESSARY)
		
		
		pdb.gimp_image_rotate(image, 2)
		pattern_layer_lr = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"Texture layer_2",100,0)

		pdb.gimp_image_insert_layer(image,pattern_layer_lr,None,0)
		pdb.gimp_selection_invert(image)
		pdb.gimp_image_select_item(image, 2, new_vectors_br)
		
		pdb.gimp_context_get_pattern()
		pdb.gimp_drawable_fill(pattern_layer_lr, 4)
		pdb.gimp_selection_invert(image)
		pdb.gimp_edit_clear(pattern_layer_lr)
		pdb.gimp_selection_none(image)		
		layer_copy_h = pdb.gimp_layer_new_from_drawable(pattern_layer_lr, image)
		pdb.gimp_image_insert_layer(image,layer_copy_h,None,0)
		pdb.gimp_item_transform_flip_simple(layer_copy_h, 1, 1, 0)

		x=pdb.gimp_image_merge_down(image,layer_copy_h,EXPAND_AS_NECESSARY)
		
		if c_type==1:
			pdb.gimp_image_raise_layer(image, mer)
			pdb.gimp_image_set_active_layer(image,mer)
			x=pdb.gimp_image_merge_down(image,mer,EXPAND_AS_NECESSARY)
				
		x=pdb.gimp_image_merge_down(image,x,EXPAND_AS_NECESSARY)			
		pdb.gimp_image_rotate(image, 0)	
		e_layer=pdb.gimp_layer_new_from_drawable(x, image)
		pdb.gimp_image_insert_layer(image,e_layer,None,0)		
		pdb.gimp_selection_none(image)
		
	if sel_col==5 and use_shape!=1:
		new_layer = e_layer
		gimp.gradient_get_uniform_samples(gimp.context_get_gradient(),4,reverse_grad)

		pdb.gimp_image_select_item(image, 2, new_vectors_bl)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, e, d,)	#d=b+e	
		else:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, e, d,)

		pdb.gimp_image_select_item(image, 2, new_vectors_br)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, wi+e-b, d, wi+e, d,)
		else:		
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, wi+d, d, wi+d+b, d,)
			
		pdb.gimp_image_select_item(image, 2, new_vectors_bt)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, d, e,)
		else:		
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, d, e,)
			
		pdb.gimp_image_select_item(image, 2, new_vectors_bb)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, wi+e, hi+e-b, wi+e, hi+e)	
		else:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, hi+d , d, hi+d+b)
	
	
	if sel_col==6:	

		new_layer = e_layer	
		pdb.gimp_selection_none(image)
		pdb.gimp_item_set_visible(layer, 1)		
		perspective_layer = pdb.gimp_layer_new_from_drawable(layer, image)
		pdb.gimp_image_insert_layer(image, perspective_layer, None, 0)
		pdb.gimp_item_set_name(perspective_layer, "Perspective")
		if scale_p!=0:
			pdb.gimp_drawable_transform_scale(perspective_layer, -2*d, -2*d, image.width+2*d,image.height+2*d, 0,1,0,1,1)
		#pdb.gimp_image_lower_layer_to_bottom(image, perspective_layer)

		pdb.gimp_image_select_item(image, 2, perspective_layer)
		
		lb_layer = pdb.gimp_layer_copy(perspective_layer,TRUE)
		pdb.gimp_image_insert_layer(image,lb_layer,None,0)
		pdb.gimp_selection_none(image)
		if scale_p==0:
			pdb.gimp_item_transform_perspective(lb_layer, e,e, d,d, e,hi+e, d,hi+e-b)#d, d, d,hi+e-b,  e,hi+e, e,e,
		else:		
			pdb.gimp_item_transform_perspective(lb_layer,e,e, d,d, e,hi + 2*d-e, d,hi + 2*d-d)
		#pdb.gimp_edit_paste(lb_layer ,0)

		rb_layer = pdb.gimp_layer_copy(perspective_layer,TRUE)
		pdb.gimp_image_insert_layer(image,rb_layer,None,0)
		if scale_p==0:
			pdb.gimp_item_transform_perspective(rb_layer, wi+e-b,d, wi+e,e, wi+e-b,hi+e-b, wi+e,hi+e,) # 
		else:	
			pdb.gimp_item_transform_perspective(rb_layer, wi+d,d,  wi+d+b,e,  wi+d,hi+d,  wi+d+b,hi+d+b,)
		merged_layer = pdb.gimp_image_merge_down(image,rb_layer,EXPAND_AS_NECESSARY)
		
		
		tb_layer = pdb.gimp_layer_copy(perspective_layer,TRUE)
		pdb.gimp_image_insert_layer(image,tb_layer,None,0)
		if scale_p==0:
			pdb.gimp_item_transform_perspective(tb_layer,e,e,wi+e,e, d,d,  wi+e-b,d)
		else:
			pdb.gimp_item_transform_perspective(tb_layer,e,e,wi+d+b,e, d,d,  wi+d,d)
		merged_layer = pdb.gimp_image_merge_down(image,tb_layer,EXPAND_AS_NECESSARY)

		bb_layer = pdb.gimp_layer_copy(perspective_layer,TRUE)
		pdb.gimp_image_insert_layer(image,bb_layer,None,0)
		if scale_p==0:
			pdb.gimp_item_transform_perspective(bb_layer,d,hi+e-b ,wi+e-b,hi+e-b, e,hi+e, wi+e,hi+e, )
		else:
			pdb.gimp_item_transform_perspective(bb_layer, d,hi+d, wi+d,hi+d,  e,hi+d+b,   wi+d+b,hi+d+b,)
		merged_layer = pdb.gimp_image_merge_down(image,bb_layer,EXPAND_AS_NECESSARY)


		if scale_p>0:
			pdb.gimp_item_set_visible(perspective_layer, 0)		
		border_p=pdb.gimp_layer_new_from_drawable(merged_layer, image)
		pdb.gimp_image_insert_layer(image, border_p, None, 0)
		pdb.gimp_image_lower_item(image, border_p)
		if pp>0:
			pdb.gimp_image_raise_item(image, passe_partout)
		pdb.gimp_selection_none(image)
		e_layer=merged_layer
		lm=merged_layer		
	if c_type==2 and sel_col==5:

		gimp.gradient_get_uniform_samples(gimp.context_get_gradient(),4,reverse_grad)
		pdb.gimp_image_select_item(image, 2, new_vectors_bl)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, e, d,)
		else:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d, d, e, d,)

		pdb.gimp_image_select_item(image, 2, new_vectors_br)
		if scale_p==0:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, wi-b, d, wi, d,)
		else:
			pdb.gimp_edit_blend(new_layer, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, d+wi, d, wi+d+b, d,)		
	pdb.gimp_image_remove_layer(image, copy)	

	new_vectors_f = pdb.gimp_vectors_new(image,"FRAME_IN")
	pdb.gimp_image_insert_vectors(image, new_vectors_f, None, -1)
	if scale_p==0:
		pdb.gimp_selection_none(image)
		pdb.gimp_vectors_stroke_new_from_points(new_vectors_f, 0, 30, newline( d,d,  wi+e-b,d,  wi+e-b,hi+e-b,  d,hi+e-b,  d,d), 1)		
	else:
		pdb.gimp_vectors_stroke_new_from_points(new_vectors_f, 0, 30, newline( d,d,  wi+d,d,  wi+d,hi+d,  d,hi+d,  d,d), 1)	
	if use_shape==1:
		sh_in=round_corners(image, new_vectors_f, 0, distance, shape_option, 0, 0)
		pdb.gimp_item_set_visible(sh_in, False)
		pdb.gimp_image_select_item(image, 2, sh_in)		
		pdb.gimp_edit_clear(background_layer)
	if use_shape==0 or use_shape==2:	
		pdb.gimp_image_select_item(image, 2, new_vectors_f)
		pdb.gimp_edit_clear(background_layer)
	

	pdb.gimp_edit_clear(e_layer)
	
	new_vectors_e = pdb.gimp_vectors_new(image,"FRAME_OUT")
	pdb.gimp_image_insert_vectors(image, new_vectors_e, None, -1)
	if scale_p==0:
		pdb.gimp_selection_none(image)
		pdb.gimp_vectors_stroke_new_from_points(new_vectors_e, 0, 30, newline( e,e,  wi+e,e,  wi+e,hi+e,  e,hi+e,  e,e), 1)	
	if scale_p>0:
		pdb.gimp_vectors_stroke_new_from_points(new_vectors_e, 0, 30, newline(e,e,  wi+2*d-e,e,  wi+2*d-e,hi+2*d-e,  e,hi+2*d-e,  e,e), 1)
	
	if use_shape>0:
		sh_out=round_corners(image, new_vectors_e, 0, distance, shape_option, 0, 0)
		pdb.gimp_item_set_visible(sh_out, False)
		pdb.gimp_image_select_item(image, 2, sh_out)	
	
	if use_shape==0 or use_shape==2:
		pdb.gimp_item_set_visible(new_vectors_e, False)
		pdb.gimp_image_select_item(image, 2, new_vectors_e)
	
	if e>0 and sel_col !=4 and sel_col !=7:
		pdb.gimp_selection_invert(image)
		pdb.gimp_item_set_name(e_layer, "Border sel color")
		pdb.gimp_edit_clear(e_layer)		
		#pdb.gimp_selection_none(image)
	new_layer = e_layer
	if sel_col==4 and use_shape==1:

		pattern_layer_bt = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"Pattern layer",100,0)
		pdb.gimp_image_insert_layer(image,pattern_layer_bt,None,0)
		pdb.gimp_context_get_pattern()
		pdb.gimp_drawable_fill(pattern_layer_bt, 4)
		pdb.gimp_image_select_item(image, 2, sh_out)
		pdb.gimp_selection_invert(image)
		pdb.gimp_edit_clear(pattern_layer_bt)
		pdb.gimp_image_select_item(image, 2, sh_in)
		pdb.gimp_edit_clear(pattern_layer_bt)
		
		e_layer=pdb.gimp_layer_new_from_drawable(pattern_layer_bt, image)
		pdb.gimp_image_insert_layer(image,e_layer,None,0)		
		pdb.gimp_selection_none(image)

		new_layer=pattern_layer_bt
	if sel_col==5 and use_shape==1:
		pdb.gimp_selection_none(image)
		gimp.gradient_get_uniform_samples(gimp.context_get_gradient(),4,reverse_grad)
		
		gradient_layer_bt = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"Gradient layer",100,0)
		pdb.gimp_image_insert_layer(image,gradient_layer_bt,None,0)
		pdb.gimp_edit_blend(gradient_layer_bt, 3, 0, gradtype, 100, 0, 0, reverse_grad, 0, 1, 4, 0, 0, 0 , wi, hi)


		pdb.gimp_image_select_item(image, 2, sh_out)
		pdb.gimp_selection_invert(image)
		pdb.gimp_edit_clear(gradient_layer_bt)
		pdb.gimp_image_select_item(image, 2, sh_in)
		pdb.gimp_edit_clear(gradient_layer_bt)
		
		e_layer=pdb.gimp_layer_new_from_drawable(gradient_layer_bt, image)
		pdb.gimp_image_insert_layer(image,e_layer,None,0)		
		pdb.gimp_selection_none(image)

		new_layer=gradient_layer_bt	
	if efect==0: # Do not use 
		border_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, border_layer, None, 0)
		pdb.gimp_item_set_name(border_layer, "Border layer-mode")
		lm = border_layer
		#lm = e_layer
		
	if efect==1:	# apply_canvas
		
		pdb.gimp_image_select_item(image, 2, new_layer)
		apply_canvas_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, apply_canvas_layer, None, 0)
		pdb.gimp_item_set_name(apply_canvas_layer, "Apply_canvas")		
		pdb.plug_in_apply_canvas(image, apply_canvas_layer, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)		
		lm = apply_canvas_layer
		
	if efect==2:	# bump_map
		
		pdb.gimp_image_select_item(image, 2, new_layer)
		bump_map = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, bump_map, None, 0)
		pdb.gimp_item_set_name(bump_map, "Bump Map")		
		pdb.plug_in_bump_map(image, e_layer, e_layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)		
		lm = bump_map		
		
	if efect==3:	# cubism
		pdb.gimp_image_select_item(image, 2, e_layer)
		cubism_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, cubism_layer, None, 0)
		pdb.gimp_item_set_name(cubism_layer, "Cubism")
		pdb.plug_in_cubism(image, cubism_layer, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)	
		lm = cubism_layer
		
	if efect==4:	# desaturate
		pdb.gimp_image_select_item(image, 2, e_layer)
		desaturate = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, desaturate, None, 0)
		pdb.gimp_item_set_name(desaturate, "Desaturate")
		pdb.gimp_drawable_desaturate(desaturate, 0)
		pdb.gimp_selection_none(image)	
		lm = desaturate		
		
	if efect==5:	# emboss
		pdb.gimp_image_select_item(image, 2, new_layer)
		emboss_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, emboss_layer, None, 0)
		pdb.gimp_item_set_name(emboss_layer, "Emboss")
		pdb.plug_in_emboss(image, emboss_layer, 1, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)	
		lm = emboss_layer
		
	if efect==6:	# engrawe
		pdb.gimp_image_select_item(image, 2, new_layer)
		engrawe_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, engrawe_layer, None, 0)
		pdb.gimp_item_set_name(engrawe_layer, "Engrawe")
		pdb.plug_in_engrave(image, engrawe_layer, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)		
		lm = engrawe_layer
		
	if efect==7:	# FELIMAGE
		pdb.gimp_image_select_item(image, 2, new_layer)
		fimg_noise_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, fimg_noise_layer, None, 0)
		pdb.gimp_item_set_name(fimg_noise_layer, "Fimg_noise")		
		pdb.plug_in_fimg_noise(image, fimg_noise_layer,"", run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)			
		lm = fimg_noise_layer
		
	if efect==8:	# gimpressionist
		pdb.gimp_image_select_item(image, 2, new_layer)
		gimpressionist_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, gimpressionist_layer, None, 0)
		pdb.gimp_item_set_name(gimpressionist_layer, "Gimpressionist")
		pdb.plug_in_gimpressionist(image, gimpressionist_layer,"", run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)	
		lm = gimpressionist_layer
			
	if efect==9:	# mosaic
		pdb.gimp_image_select_item(image, 2, new_layer)
		mosaic_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, mosaic_layer, None, 0)
		pdb.gimp_item_set_name(mosaic_layer, "Mosaic")		
		pdb.plug_in_mosaic(image, mosaic_layer, 10, 1, 4, 1, 0, 0, 0, 1, 0, 0, 0, 1,run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)
		lm = mosaic_layer
		
	if efect==10:	# NEWSPRINT
		pdb.gimp_image_select_item(image, 2, new_layer)
		newsprint_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, newsprint_layer, None, 0)
		pdb.gimp_item_set_name(newsprint_layer, "Newsprint")			
		pdb.plug_in_newsprint(image, newsprint_layer, 1, 3, 50, 180, 0, 1.0, 0, 90.0, 0, 5.0, 0, 10,run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)				
		lm = newsprint_layer
		
	if efect==11:	# NOISE GENERATOR
		pdb.gimp_image_select_item(image, 2, new_layer)
		noise_generator_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, noise_generator_layer, None, 0)
		pdb.gimp_item_set_name(noise_generator_layer, "Noise_generator")		
		pdb.plug_in_noise_generator(image, noise_generator_layer, 1, 1, 1, 1, 1, 1,run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)
		lm = noise_generator_layer
		
	if efect==12:	# pixelize
		pdb.gimp_image_select_item(image, 2, new_layer)
		pixelize_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, pixelize_layer, None, 0)
		pdb.gimp_item_set_name(pixelize_layer, "Pixelize")			
		pdb.plug_in_pixelize2(image, pixelize_layer, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)
		lm = pixelize_layer

	if efect==13:	# ripple
		pdb.gimp_image_select_item(image, 2, new_layer)
		ripple_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, ripple_layer, None, 0)
		pdb.gimp_item_set_name(ripple_layer, "Ripple")		
		pdb.plug_in_ripple(image, ripple_layer, 1, 1, 1, 1, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)	
		lm = ripple_layer
		
	if efect==14:	# spread
		pdb.gimp_image_select_item(image, 2, new_layer)
		spread_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, spread_layer, None, 0)
		pdb.gimp_item_set_name(spread_layer, "Spread")		
		pdb.plug_in_spread(image, spread_layer, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)		
		lm = spread_layer
		
	if efect==15:	# video
		pdb.gimp_image_select_item(image, 2, new_layer)
		video_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, video_layer, None, 0)
		pdb.gimp_item_set_name(video_layer, "Video")		
		pdb.plug_in_video(image, video_layer, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)
		lm = video_layer
		
	if efect==16:	# VORONOI
		pdb.gimp_image_select_item(image, 2, new_layer)
		voronoi_layer = pdb.gimp_layer_new_from_drawable(new_layer, image)
		pdb.gimp_image_insert_layer(image, voronoi_layer, None, 0)
		pdb.gimp_item_set_name(voronoi_layer, "Voronoi")		
		pdb.plug_in_voronoi(image, voronoi_layer, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, run_mode=RUN_INTERACTIVE)
		pdb.gimp_selection_none(image)		
		lm = voronoi_layer
		
	if modepat==0: # "LAYER-MODE-NORMAL 28"
		pdb.gimp_layer_set_mode(lm, 28)
	if modepat==1:  # "LAYER-MODE-DISSOLVE 1"
		pdb.gimp_layer_set_mode(lm, 1)
	if modepat==2: # "LAYER-MODE-OVERLAY 23"
		pdb.gimp_layer_set_mode(lm, 23)
	if modepat==3:	# "LAYER-MODE-LCH-HUE 24"
		pdb.gimp_layer_set_mode(lm, 24)
	if modepat==4:	# "LAYER-MODE-LCH-CHROMA 25"
		pdb.gimp_layer_set_mode(lm, 25)
	if modepat==5:	# "LAYER-MODE-LCH-COLOR 26"
		pdb.gimp_layer_set_mode(lm, 26)
	if modepat==6:	 # "LAYER-MODE-LCH-LIGHTNESS 27"
		pdb.gimp_layer_set_mode(lm, 27)
	#if modepat==7: # "LAYER-MODE-BEHIND 29"
		#pdb.gimp_layer_set_mode(lm, 29)
	if modepat==7:	# "LAYER-MODE-MULTIPLY 30"
		pdb.gimp_layer_set_mode(lm, 30)
	if modepat==8:	# "LAYER-MODE-SCREEN 31"
		pdb.gimp_layer_set_mode(lm, 31)
	if modepat==9:	# "LAYER-MODE-DIFFERENCE 32"
		pdb.gimp_layer_set_mode(lm, 32)
	if modepat==10:	 # "LAYER-MODE-ADDITION 33"
		pdb.gimp_layer_set_mode(lm, 33)
	if modepat==11:	 # "LAYER-MODE-SUBTRACT 34"
		pdb.gimp_layer_set_mode(lm, 34)
	if modepat==12:	# "LAYER-MODE-DARKEN-ONLY 35"
		pdb.gimp_layer_set_mode(lm, 35)
	if modepat==13:	# "LAYER-MODE-LIGHTEN-ONLY 36"
		pdb.gimp_layer_set_mode(lm, 36)
	if modepat==14:	# "LAYER-MODE-HSV-HUE 37"
		pdb.gimp_layer_set_mode(lm, 37)
	if modepat==15:	 # "LAYER-MODE-HSV-SATURATION 38"
		pdb.gimp_layer_set_mode(lm, 38)
	if modepat==16:	# "LAYER-MODE-HSL-COLOR 39"
		pdb.gimp_layer_set_mode(lm, 39)
	if modepat==17:	# "LAYER-MODE-HSV-VALUE 40"	
		pdb.gimp_layer_set_mode(lm, 40)
	if modepat==18:	 # "LAYER-MODE-DIVIDE 41"
		pdb.gimp_layer_set_mode(lm, 41)
	if modepat==19:	 # "LAYER-MODE-DODGE 42"
		pdb.gimp_layer_set_mode(lm, 42)
	if modepat==20:	 # "LAYER-MODE-BURN 43"
		pdb.gimp_layer_set_mode(lm, 43)
	if modepat==21:	# "LAYER-MODE-HARDLIGHT 44"
		pdb.gimp_layer_set_mode(lm, 44)
	if modepat==22:	# "LAYER-MODE-SOFTLIGHT 45"
		pdb.gimp_layer_set_mode(lm, 45)
	if modepat==23:	# "LAYER-MODE-GRAIN-EXTRACT 46"
		pdb.gimp_layer_set_mode(lm, 46)
	if modepat==24:	# "LAYER-MODE-GRAIN-MERGE 47"
		pdb.gimp_layer_set_mode(lm, 47)
	if modepat==25:	# "LAYER-MODE-VIVID-LIGHT 48"
		pdb.gimp_layer_set_mode(lm, 48)
	if modepat==26:	# PIN-LIGHT (49)
		pdb.gimp_layer_set_mode(lm, 49)
	if modepat==27:	# LINEAR-LIGHT (50)
		pdb.gimp_layer_set_mode(lm, 50)
	if modepat==28:	# HARD-MIX (51)
		pdb.gimp_layer_set_mode(lm, 51)
	if modepat==29:	# EXCLUSION (52)
		pdb.gimp_layer_set_mode(lm, 52)
	if modepat==30:	# LINEAR-BURN (53)
		pdb.gimp_layer_set_mode(lm, 53)
	if modepat==31:	# LUMA-DARKEN-ONLY (54)
		pdb.gimp_layer_set_mode(lm, 54)
	if modepat==32:	# LUMA-LIGHTEN-ONLY (55)
		pdb.gimp_layer_set_mode(lm, 55)
	if modepat==33:	# LUMINANCE (56)
		pdb.gimp_layer_set_mode(lm, 56)
		
	pdb.gimp_layer_set_opacity(lm, opacity)	
	pdb.gimp_image_raise_layer_to_top(image, lm)

	if stroke_out==1 or stroke_in==1 or stroke_layer==1:
		stroke = pdb.gimp_layer_new(image,image.width,image.height,RGBA_IMAGE,"box stroke",100,0)
		pdb.gimp_image_insert_layer(image,stroke,None,0)
	
	if stroke_out==1:
		gimp.set_foreground(color_out)		
		pdb.gimp_selection_none(image)
		if use_shape==1:
			pdb.gimp_drawable_edit_stroke_item(stroke ,sh_out)
		if use_shape==0 or use_shape==2:
			pdb.gimp_drawable_edit_stroke_item(stroke ,new_vectors_e)
		
	if stroke_in==1:
		gimp.set_foreground(color_in)
		pdb.gimp_selection_none(image)
		if use_shape==1:
			pdb.gimp_drawable_edit_stroke_item(stroke ,sh_in)
		if use_shape==0 or use_shape==2:
			pdb.gimp_drawable_edit_stroke_item(stroke ,new_vectors_f)

	if pp>0:	
		new_vectors_layer = pdb.gimp_vectors_new(image,"Visibly_Layer")
		pdb.gimp_image_insert_vectors(image, new_vectors_layer, None, -1)
		if scale_p==0:
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_layer, 0, 30, newline( d+pp,d+pp,  wi+e-b-pp,d+pp,  wi+e-b-pp,hi+e-b-pp,  d+pp,hi+e-b-pp,  d+pp,d+pp), 1)
		else:																		
			pdb.gimp_vectors_stroke_new_from_points(new_vectors_layer, 0, 30, newline( d+pp,d+pp,  wi+d-pp,d+pp,  wi+d-pp,hi+d-pp,  d+pp,hi+d-pp,  d+pp,d+pp), 1)
			
		if use_shape==1 or use_shape==2:
			shape_vectors_layer=round_corners(image, new_vectors_layer, 0, distance, shape_option, 0, 0)			
			pdb.gimp_item_set_visible(shape_vectors_layer, False)
			pdb.gimp_image_select_item(image, 2, shape_vectors_layer)
			#
			
		else:	
			pdb.gimp_item_set_visible(new_vectors_layer, False)
			pdb.gimp_image_select_item(image, 2, new_vectors_layer)		
		pdb.gimp_edit_clear(passe_partout)

		if use_shape==1:
			pdb.gimp_image_select_item(image, 2, sh_out)		
			pdb.gimp_selection_invert(image)
			pdb.gimp_edit_clear(passe_partout)

			
		pdb.gimp_layer_resize_to_image_size(passe_partout)
		pdb.gimp_selection_none(image)
		
	if stroke_layer==1 and pp>0:
		gimp.set_foreground(color_sl)
		pdb.gimp_selection_none(image)
		if use_shape!=0:
			pdb.gimp_drawable_edit_stroke_item(stroke ,shape_vectors_layer)
		else:		
			pdb.gimp_drawable_edit_stroke_item(stroke ,new_vectors_layer)		

	if use_shape!=0 and pp>0:

		pdb.gimp_image_select_item(image, 2, shape_vectors_layer)	
		pdb.gimp_selection_invert(image)
		pdb.gimp_edit_clear(active_layer)
		pdb.gimp_selection_none(image)

			
	if drop_shadow:

		pdb.gimp_layer_resize_to_image_size(e_layer)
		color_to_gegl=rgb_to_hex(color_s)
		
		if grow_shape==0:
			grow_s='square'
		if grow_shape==1:
			grow_s='circle'
		if grow_shape==2:
			grow_s='diamond'
		if drop_shadow==1 or drop_shadow==3:			
			gegl_command_preview(image, e_layer, "dropshadow x=%d y=%d radius=%d grow-shape=%s grow-radius=%d color=%s opacity=%d" % (shadow_offset_x, shadow_offset_y, shadow_blur,grow_s,grow_radius, color_to_gegl, shadow_opacity))

		pdb.gimp_image_raise_layer_to_top(image, lm)
		if pp>0:	
			if stroke_out==1 or stroke_in==1 or stroke_layer==1:
				pdb.gimp_image_raise_layer(image, e_layer)
				pdb.gimp_image_raise_layer_to_top(image, stroke)
			if drop_shadow==2 or drop_shadow==3:			
				gegl_command_preview(image, passe_partout, "dropshadow x=%d y=%d radius=%d grow-shape=%s grow-radius=%d color=%s opacity=%d" % (shadow_offset_x, shadow_offset_y, shadow_blur,grow_s,grow_radius, color_to_gegl, shadow_opacity))
				passe_partout.name = "Drop Shadow Passe-partout"
			if scale_p!=0 or sel_col==6:	
				pdb.gimp_image_raise_layer(image, passe_partout)
				
			pdb.gimp_image_select_item(image, 2, lm)
			e_layer.name = "Drop Shadow Frame"
			pdb.gimp_edit_clear(e_layer)
	if use_distress==1:

		distress(image, active_layer, select+d+pp,threshold,spread,granularity,smooth,color_d)
		if sel_col==6 and scale_p!=0:
			pdb.gimp_item_set_visible(layer, 0)
		if sel_col==6 and scale_p==0:
			pdb.gimp_item_set_visible(layer, 0)
			pdb.gimp_item_set_visible(perspective_layer, 0)			
		if use_shape!=0 and background_col==3:
			if use_shape==1:
				pdb.gimp_image_select_item(image, 2, sh_in)
			if use_shape==0 or use_shape==2:
				pdb.gimp_image_select_item(image, 2, new_vectors_f)
			
			pdb.gimp_selection_invert(image)
			layer_list = image.layers
			pdb.gimp_edit_clear(layer_list[-1])		

	if scale_p==2:	# Add to layer & Scale to original
		pdb.gimp_image_scale(image, wi-2*pp, hi-2*pp)
	
	if del_path==1:
		for v in image.vectors:
			pdb.gimp_image_remove_vectors(image, v)		
	if merge==1:
		pdb.gimp_image_merge_visible_layers(image, 1)
		
	pdb.gimp_selection_none(image)
	
#######################################################################################
	# add for "Preview"		
	view = pdb.gimp_layer_new_from_visible(image, image, "Preview")
	pdb.gimp_image_insert_layer(image, view, None, 0)
#######################################################################################

	pdb.gimp_image_undo_group_end(image)
	pdb.gimp_context_pop()	


hlp = "Create border step by step for GIMP 2.10.\nOnly works for RGB images with an Alpha Channel.\nRecommended interesting use:\n - select gradient FG-to-BG-(RGB) that is: BG "u'\u21A4'" FG "u'\u21A6'" BG,\n - repeat the plugin for 'Swap Colors'\n - if Merge Layers = No, You can select the border layer mode experimentally.\nFor proper operation (Additional effect) it is necessary to install additional plugins.\nAdditional effect mostly it works interactively (it brings up another plugin window).\nFor Windows10 and Gimp-2.10, all additional plugins are available on:\n<b>https://www.gimpscripts.net/</b>  "	\
+ "\n\nResizing method 'Preview':\n1. Disable preview.\n2. Drag the corner.\n3. Turn on the preview again.\n\nOnly the size of the initial (active) layer is displayed, so the preview\nmay show an image slightly different from the final result.\nUsing options:'Scale to original layer size' helps with the preview." \
+ "\nIf You change a lot of parameters, it is better to turn off the preview for this time." \
+ "\nThe plugin uses code snippets created by:\n<b>Claude Lion [cli345], J.F. Garcia [arakne],\nTin Tran [Trandoductin], Markku Koppinen [aka: Ottia Tuota].</b>\nThank You very much for sharing.\n\nI wish You good use of this plugin:\n		MareroQ.\nHave fun."
register (
	"MrQ_BorderBox",
	"BorderBox Step by Step Ver.1.5",
	"Create border step by step for GIMP 2.10\nVersion 1.5",
	"MareroQ",
	"GPL",
	"2021",
	"BorderBox... ",
	"RGBA",
	[
	(PF_IMAGE,    "image", "", None),
	(PF_DRAWABLE, "drawable", "", None),
	(PF_SPINNER,  "background", "Underlay:", 10, (0, 1000, 1),{'newLine':'0'}),
	(PF_OPTION,   "background_col", "(px)		Fill:", 2, [ "FOREGROUND", "BACKGROUND", "WHITE", "TRANSPARENT", "ACTIVE-PATTERN", "BLURRED Layer"],{'newLine':1}),	
	(PF_SPINNER,  "frame", "Frame: ", 20, (0, 1000, 1),{'newLine':'0'}),
	(PF_OPTION,   "sel_col", "(px)		Fill:", 5, [ " Foreground Color "," Background Color "," Black "," White "," Active Pattern "," Active Gradient " ," PERSPECTIVE "," Do not use "],{'newLine':1}),		
	(PF_SPINNER,  "pp", "Passe\n-partout:\n(Mat)", 30, (0, 1000, 1),{'newLine':'0'}),
	(PF_OPTION,   "pp_col", "(px)		Fill:", 2, [ "FOREGROUND", "BACKGROUND", "WHITE", "TRANSPARENT", "ACTIVE-PATTERN","ACTIVE-PATTERN\nDesaturate"],{'newLine':1}),	
	(PF_OPTION,   "c_type", "Corner type:", 0, ["SLANTED","RECTANGULAR T-B","RECTANGULAR R-L"],{'newLine':'0'}),
	(PF_OPTION,   "gradient_type", "	Gradient\n	type:", 0, ["LINEAR","SHAPEBURST-ANGULAR","SHAPEBURST-SPHERICAL","SHAPEBURST-DIMPLED"],{'newLine':1}),	
	(PF_COLOR,    "color_fg", "Color:	FG", (50, 25, 225),0,{'newLine':'0'}),
	(PF_COLOR,    "color_bg", "		BG", (50, 250, 225),0,{'newLine':1}),
	(PF_OPTION,   "swap", "	Swap\n	Color: ", 0, [" No "," Yes (replace FG with BG)"],{'newLine':'0','iniCol':2}),	

	(PF_OPTION,   "efect", "Additional\neffect:\n(interact.) ", 0, [" DON'T USE "," APPLY CANVAS "," BUMP MAP"," CUBISM ", " DESATURATE "," EMBOSS "," ENGRAVE ",
	" FELIMAGE "," GIMPRESSIONIST "," MOSAIC "," NEWSPRINT "," NOISE GENERATOR", " PIXELIZE "," RIPPLE "," SPREAD "," VIDEO "," VORONOI "]),
	(PF_OPTION, "modepat", "Layer-mode\n(for effect)", 0, 
	["NORMAL","DISSOLVE","OVERLAY", "LCH-HUE", "LCH-CHROMA", "LCH-COLOR", "LCH-LIGHTNESS", "MULTIPLY", "SCREEN", "DIFFERENCE", "ADDITION", "SUBTRACT", "DARKEN-ONLY",
	"LIGHTEN-ONLY", "HSV-HUE", "HSV-SATURATION", "HSL-COLOR", "HSV-VALUE", "DIVIDE", "DODGE", "BURN", "HARDLIGHT", "SOFTLIGHT", "GRAIN-EXTRACT", "GRAIN-MERGE", "VIVID-LIGHT",
	"PIN-LIGHT", "LINEAR-LIGHT", "HARD-MIX", "EXCLUSION", "LINEAR-BURN", "LUMA-DARKEN-ONLY", "LUMA-LIGHTEN-ONLY", "LUMINANCE"],{'newLine':'0'}),
	(PF_SPINNER,  "opacity", "	Opacity: ", 100, (0, 100, 1),{'newLine':'1'}),	
	(PF_OPTION, "scale_p", "Apply as: ", 0, [" On a layer "," Add to layer ", " Add to layer & Scale\n to original size"],{'newLine':'0'}),
	(PF_OPTION, "merge", "	Merge\n	layers: ", 0, [" No "," Yes "],{'newLine':'1'}),

	(PF_OPTION, "use_shape", "Use Shape:", 0, [" DON'T USE "," For Frame & Passe-Partout ", " Only for Passe-Partout"],{'tab':1}),
	(PF_OPTION,  "shape_option","Round Corners:", 1, [" Circular ", " Straight ", " Parabola  "," Close to vertex "," Through vertex "," Cusp "],{'tab':1}),
	(PF_SPINNER, "distance", "Distance:", 50, (1, 1000, 1),{'tab':1}),

	(PF_TITLE , "stroke_tx"  , "\n<b>				Stroke With ACTIVE Brush:</b>\n				  (Paintbrush, Pencil, MyPaint)",(0.0,0.0), None  ,{'cols':26,'iniCol':1,'newLine':1,'tab':2}),	
	(PF_OPTION, "stroke_out", "Outside frame box:", 0, [" No "," Yes "],{'newLine':1,'tab':2}),
	(PF_COLOR,   "color_out", "Color:", (0, 0, 0),0,{'tab':2}),	
	(PF_OPTION, "stroke_in", "Inside frame box:", 0, [" No "," Yes "],{'newLine':1,'tab':2}),
	(PF_COLOR,   "color_in", "Color:", (255, 255, 255),0,{'tab':2}),	
	(PF_OPTION, "stroke_layer", "Visible Layer:", 0, [" No "," Yes "],{'newLine':1,'tab':2}),
	(PF_COLOR,   "color_sl", "Color:", (0, 0, 0),0,{'tab':2}),	
	(PF_RADIO,  "del_path", "Delete paths: ", 0, (("Yes", 1),("No",0)),{'newLine':1,'tab':2}),
	
	(PF_OPTION,  "drop_shadow","Add Drop Shadow:", 1, [" No "," To Border ", " To Passe-Partout ", " To Both  "],{'tab':3}),	
	(PF_SPINNER, "shadow_offset_x", "Shadow Offset X:", 5, (-1000, 1000, 1),{'tab':3}),
	(PF_SPINNER, "shadow_offset_y", "Shadow Offset Y:", 5, (-1000, 1000, 1),{'tab':3}),
	(PF_SPINNER, "shadow_blur", "Shadow Blur Radius:", 5, (0, 300, 1),{'tab':3}),
	(PF_OPTION,  "grow_shape", "Grow Shape: ", 0, [" Square "," Circle ", " Diamond "],{'tab':3}),	
	(PF_SPINNER, "grow_radius", "Grow Radius:", 5, (-100, 100, 1),{'tab':3}),
	(PF_COLOR,   "color_s", "Color:", (0, 0, 0),0,{'tab':3}),
	(PF_SPINNER, "shadow_opacity", "Shadow Opacity:", 1.5, (0, 2, 0.01),{'tab':3}),
	

	(PF_OPTION, "use_distress", "Use Distress:", 0, [" DON'T USE "," YES "],{'tab':4}),	
	(PF_SPINNER, "select", 'Distanse Select', 6, (1,100,1),{'tab':4}),	
	(PF_SPINNER, "threshold", " Threshold (bigger 1<-->254 smaller):", 127, ( 1, 254, 1),{'tab':4}),
	(PF_SPINNER, "spread", " Spread:" , 8, ( 0, 200, 1),{'tab':4}),
	(PF_SPINNER, "granularity", " Granularity (1 is low):", 4, ( 1, 20, 1),{'tab':4}),
	(PF_SPINNER, "smooth", " Smooth:", 2, ( 1, 150, 1),{'tab':4}), 
	(PF_COLOR,   "color_d", "Filling the surroundings:", (255, 255, 255),0,{'tab':4}),	
	(PF_TITLE  , "help"  , hlp ,(10,10), None           ,{'tab':5}),		
	(PF_TAB    , "ptab"    , "tabs" , 0, (' Frames ','Round Corners ' ,' Stroke Paths ', " Shadows ","Distress", "Help "),{'newLine':'1','tab':4} ),	
    ],
    [],
    MrQ_BorderBox, menu = "<Image>/Filters/Decor"
    )

main()
